/**
 * 
 */
package eos.agents;

import java.util.Map;

import eos.framework.Economy;
import eos.framework.Good;
import eos.models.Baseline2Model;
import eos.models.Laborer;
import eos.models.Baseline2Model.Baseline2Laborer;
import eos.models.SingleOwnerFirms.SingleOwnerFirm;

/**
 * @author Michael Adelson
 * 
 */
public class BL2SeekingOwnerLaborer extends Baseline2Laborer {

    /**
     * A comfortable threshold of food.
     */
    public static final double FOOD_THRESHOLD = 30.0 * Laborer.EAT_AMOUNT;

    /**
     * A comfortable supply of money.
     */
    public static final double MONEY_THRESHOLD = 30.0;

    /**
     * Price adjustments for buying and selling
     */
    private double foodPriceAdjustment = 0.0, laborPriceAdjustment = 0.0;

    /**
     * Construct a BL2SeekingLaborer
     * 
     * @see Baseline2Laborer#Baseline2Laborer(Economy, Map)
     */
    public BL2SeekingOwnerLaborer(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
    }

    /**
     * Buys on Food. If owns any Firms, claims a salary as necessary. Otherwise,
     * Sells labor.
     * 
     * @see eos.models.Baseline2Model.Baseline2Laborer#baseLine2LaborerSubclassAct()
     */
    protected void baseLine2LaborerSubclassAct() {
        if (this.firms == null || this.firms.isEmpty()) {
            this.bidOnLabor();
        } else {
            this.claimSalary();
        }

        this.bidOnFood();
    }

    /**
     * The last quantity of labor this Laborer tried to sell.
     */
    private double lastLaborSellQuantity;

    /**
     * Sell labor.
     */
    private void bidOnLabor() {
        // no need to sell any labor
//        if (this.money.getQuantity() >= MONEY_THRESHOLD) {
//            System.out.println("Not Selling!");
//            return;
//        }

        // initially just try to buy at historical price
        if (this.laborMarket.getLastMarketPrice() <= 0.0) {
            this.laborMarket.addSellOffer(this, Laborer.LABOR_ALLOWANCE,
                    Baseline2Model.HISTORICAL_LABOR_PRICE);
            this.lastLaborSellQuantity = Laborer.LABOR_ALLOWANCE;
            return;
        }

        // basically sold all labor, so increase prices with some probability
        if (this.lastLaborSellQuantity - this.getLastLaborSold() < 0.01 * Laborer.LABOR_ALLOWANCE) {
            if (this.economy.random() < 0.1)
                this.laborPriceAdjustment += BL2SeekingFarm.ADJUST_PERCENT
                        * this.laborMarket.getLastMarketPrice();
        } else
            // didn't sell all food, so try lowering price offer
            this.laborPriceAdjustment -= BL2SeekingFarm.ADJUST_PERCENT
                    * this.laborMarket.getLastMarketPrice();

        // finalize the price, never asking less than the price of 1-day's food
        double price = Math
                .max(
                        (this.laborMarket.getLastMarketPrice() + this.laborPriceAdjustment),
                        (this.foodMarket.getLastMarketPrice() * Laborer.EAT_AMOUNT));

        // sell an amount of labor = to money needed / price of labor
        // (at most LABOR_ALLOWANCE)
        double amountToSell = Math.max((MONEY_THRESHOLD - this.money
                .getQuantity())
                / this.laborMarket.getLastMarketPrice(),
                Laborer.LABOR_ALLOWANCE);
        if (this.food.getQuantity() < 0.75 * FOOD_THRESHOLD)
            amountToSell = Laborer.LABOR_ALLOWANCE;

        // update sell quantity
        this.lastLaborSellQuantity = amountToSell;

        // SELL!
        if (price > 0.0)
            this.laborMarket.addSellOffer(this, amountToSell, price);

        System.out.println("Selling Labor: price=" + price + " adj="
                + this.laborPriceAdjustment + " money="
                + this.money.getQuantity());
    }

    /**
     * Claim a salary from an Owned Firm
     */
    private void claimSalary() {
        if (this.money.getQuantity() < MONEY_THRESHOLD) {
            double amountPerFirm = (MONEY_THRESHOLD - this.money.getQuantity())
                    / this.firms.size();
            for (SingleOwnerFirm firm : this.firms)
                firm.requestTransfer(this.money.getName(), amountPerFirm);

            System.out.println("Owner claiming "
                    + (amountPerFirm * this.firms.size()));
        }
    }

    /**
     * Buy Food.
     */
    private void bidOnFood() {
        // can't buy anything or don't need to buy anything
        if (money.getQuantity() <= 0.0
                || this.food.getQuantity() >= FOOD_THRESHOLD)
            return;

        // initially just try to buy at historical price
        if (this.foodMarket.getLastMarketPrice() <= 0.0) {
            this.foodMarket.addBuyOffer(this, Laborer.EAT_AMOUNT,
                    Baseline2Model.HISTORICAL_FOOD_PRICE);
            return;
        }

        // bought most of the desired food, so lower prices
        // (had close to FOOD_THRESHOLD to begin the time step)
        if (this.food.getQuantity() > FOOD_THRESHOLD - 1.5 * Laborer.EAT_AMOUNT) {
            if (this.economy.random() < 1.1) // lower prices with some
                // probability
                this.foodPriceAdjustment -= BL2SeekingFarm.ADJUST_PERCENT
                        * this.foodMarket.getLastMarketPrice();
        } else
            // didn't buy all food, so try increasing price offer
            this.foodPriceAdjustment += BL2SeekingFarm.ADJUST_PERCENT
                    * this.foodMarket.getLastMarketPrice();

        // finalize the price
        double price = this.foodMarket.getLastMarketPrice()
                + this.foodPriceAdjustment;

        // buy enough food at price to restore food to FOOD_THRESHOLD
        this.foodMarket.addBuyOffer(this, FOOD_THRESHOLD
                - this.food.getQuantity(), price);

        System.out.println("Buying food: price=" + price + " adj="
                + this.foodPriceAdjustment + " food=" + this.food.getQuantity()
                + " money=" + this.money.getQuantity());
    }

}
