package eos.agents;

import java.util.*;

import eos.datatypes.SmoothedValue;
import eos.framework.*;
import eos.models.Laborer;
import static eos.models.Baseline3Model.*;

/**
 * The BudgetLaborer buys food using a budget, so as not to steadily lose money.
 * This implementation was ported to the baseline to test a non-homogenous
 * economy.
 * 
 * Author: Michael Adelson
 */
public class BL3BudgetLaborer extends Baseline3Laborer {

    /**
     * When food quantity is below the threshold the laborer bids more than the
     * market price and vice versa.
     */
    private static final double FOOD_THRESHOLD = 15;

    /**
     * Constant used in exponential smoothing.
     */
    private static final double ALPHA = 0.1;

    /**
     * An exponentially smoothed estimate of the value of labor.
     */
    private final SmoothedValue laborValue, foodValue;

    /**
     * Construct a DemandCurveLaborer in Economy with goods
     */
    public BL3BudgetLaborer(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
        this.laborValue = new SmoothedValue(0, ALPHA);
        this.foodValue = new SmoothedValue(0, ALPHA);
    }

    /**
     * Sell Labor and buy Food.
     */
    protected void baseLine3LaborerSubclassAct() {
        // update value of food
        if (this.foodMarket.getLastMarketPrice() <= 0.0) {
            this.laborValue.setValue(HISTORICAL_LABOR_PRICE);
            this.foodValue.setValue(HISTORICAL_FOOD_PRICE);
        } else {
            this.laborValue.update(this.laborMarket.getLastMarketPrice());
            this.foodValue.update(this.foodMarket.getLastMarketPrice());
        }

        bidOnLabor();
        bidOnFood();
    }

    /**
     * Sell Labor based on value as well as the BudgetLaborer's food supply.
     */
    private void bidOnLabor() {

        double laborPrice = this.laborValue.getValue();

        // factor that determines how much DCL's labor price offer changes
        if (food.getQuantity() >= FOOD_THRESHOLD)
            laborPrice *= food.getQuantity() / FOOD_THRESHOLD;
        else if (food.getQuantity() < 0.5 * FOOD_THRESHOLD) {
            if (foodMarket.getLastMarketPrice() <= 0.0)
                laborPrice = (this.laborValue.getValue() * 0.75);
            else
                // offer slightly above food price
                laborPrice = Math.min(this.laborValue.getValue(),
                        1.1 * foodMarket.getLastMarketPrice());
        }

        this.laborMarket.addSellOffer(this, LABOR_ALLOWANCE, laborPrice);
    }

    /**
     * Bid on Food in a demand curve fashion, setting a strict budget
     * constraint.
     */
    private void bidOnFood() {
        // can't buy anything with 0 money
        if (money.getQuantity() <= 0.0)
            return;

        // variables used in calculations
        double price, Qm = money.getQuantity(), Qf = food.getQuantity(), budget = Math
                .min(this.laborValue.getValue(), Qm), // default budget is value
        // of labor
        lastPrice = this.foodValue.getValue(); // foodMarket.getLastMarketPrice();

        // initial condition: buy food at half the price of labor and lower
        if (lastPrice <= 0.0) {
            this.foodMarket.addBuyOffer(this, Laborer.EAT_AMOUNT,
                    this.foodValue.getValue());
            // price = this.laborValue.getValue();
            // for (int i = 0; i < 10; i++) {
            // price /= 2.0;
            // foodMarket.addBuyOffer(this, EAT_AMOUNT, price);
            // }
            return;
        }

        // set budget
        if (Qf < EAT_AMOUNT) // desperation budget = money supply or 4 *
            // lastPrice
            budget = Math.min(4.0 * Math.max(foodMarket.getLastMarketPrice(),
                    lastPrice), Qm);
        else if (Qf < 0.5 * FOOD_THRESHOLD) // slightly higher budget than
            // normal
            budget = Math.min(1.5 * this.laborValue.getValue(), Qm);
        else if (Qf > 2.0 * FOOD_THRESHOLD) // heavy saving
            budget = budget * 0.5;
        else if (Qf > 1.5 * FOOD_THRESHOLD) // small savings
            budget = budget * 0.9;

        // make bids based on budget
        // System.out.println("budget: " + budget);
        // if desperate, try to make 2 bids above lastPrice
        if (Qf < 0.5 * FOOD_THRESHOLD) {
            if ((price = budget / (2.0 * EAT_AMOUNT)) > lastPrice) {
                foodMarket.addBuyOffer(this, 2.0 * EAT_AMOUNT, price);
            } else {
                price = Math.min(budget / EAT_AMOUNT, 1.5 * lastPrice);
                foodMarket.addBuyOffer(this, EAT_AMOUNT, price);
                budget -= price * EAT_AMOUNT;
                if (budget > 0.0) // spend remaining budget
                    foodMarket.addBuyOffer(this, EAT_AMOUNT, budget
                            / EAT_AMOUNT);
            }
        }

        // otherwise, submit a demand curve
        double slope = -lastPrice / FOOD_THRESHOLD;
        while (budget > 0.0) {
            price = 2.0 * lastPrice + Qf * slope;
            if (price <= lastPrice / 100.0)
                break; // no zero/uselessly small prices
            if (price * EAT_AMOUNT > budget)
                price = budget / EAT_AMOUNT;
            foodMarket.addBuyOffer(this, EAT_AMOUNT, price);
            budget -= price * EAT_AMOUNT;
        }
    }
}
