package eos.agents;

import java.util.Map;
import eos.models.BaselineModel.*;
import eos.framework.*;

/**
 * A very simple implementation of Laborer which tries to buy 2 food each turn.
 * Pricing for both food purchases and labor sales is based on the last price of
 * food and the SimpleLaborer2's quantity of food (relative to his
 * FOOD_THRESHOLD).
 * 
 * Author: Michael Adelson
 */
public class SimpleLaborer extends BaselineLaborer {

    /**
     * A comfortable stock of food.
     */
    private static double FOOD_THRESHOLD = 30;

    /**
     * Food and Labor Markets
     */
    private Market foodMarket, laborMarket;

    /**
     * Smoothing coefficient
     */
    public static double ALPHA = 1.0;

    /**
     * Smoothed value estimate for food.
     */
    private double value;

    /**
     * Construct a new SimpleLaborer2
     */
    public SimpleLaborer(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
        foodMarket = null;
        laborMarket = null;
    }

    /**
     * Bid on labor and food.
     */
    protected void bid() {
        if (foodMarket == null) { // initialize
            foodMarket = economy.getMarketFor(money.getName(), food.getName());
            laborMarket = economy
                    .getMarketFor(money.getName(), labor.getName());
            if (foodMarket == null || laborMarket == null)
                throw new RuntimeException(
                        "Economy must have food/money and labor/money markets");
            this.value = foodMarket.getLastMarketPrice(); // gets reset later
        } else
            value = ALPHA * foodMarket.getLastMarketPrice() + (1 - ALPHA)
                    * value;

        bidOnLabor();
        bidOnFood();
    }

    /**
     * Bid on labor.
     */
    private void bidOnLabor() {
        if (foodMarket.getLastMarketPrice() <= 0.0) {
            laborMarket.addSellOffer(this, LABOR_ALLOWANCE, 2.0 * this.economy
                    .random() + 0.1);
            return;
        }

        double price = 2.0 * value /*foodMarket.getLastMarketPrice()*/
                * (food.getQuantity() / FOOD_THRESHOLD)
                * (EAT_AMOUNT / LABOR_ALLOWANCE);
        if (price <= 0.0)
            return;
        laborMarket.addSellOffer(this, LABOR_ALLOWANCE, price);
    }

    /**
     * Bid on food.
     */
    private void bidOnFood() {
        if (foodMarket.getLastMarketPrice() <= 0.0) {
            double p1 = this.economy.random() + 0.05;
            double p2 = this.economy.random() + 0.05;
            foodMarket.addBuyOffer(this, EAT_AMOUNT, p1);
            foodMarket.addBuyOffer(this, EAT_AMOUNT, p2);
            this.value = (p1 + p2) / 2.0;
            return;
        }

        if (money.getQuantity() <= 0.0)
            return;

        double quantity = 2.0 * EAT_AMOUNT;
        double price = value // foodMarket.getLastMarketPrice()
                * (FOOD_THRESHOLD / (food.getQuantity() + 0.1));
        if ((price * quantity) > money.getQuantity())
            price = money.getQuantity() / quantity;

        if (price > 0.0)
            foodMarket.addBuyOffer(this, quantity, price);
    }

}
