package eos.io;

import java.util.*;

import eos.utilities.Strings;
import eos.framework.*;

/**
 * Can be used to track the distrubution of Goods. The output can be average or
 * total, and can be tracked by class or by total population.
 * 
 * @author Michael Adelson
 */
public class GoodTrackerDataPrinter implements DataPrinter {

    private final boolean avg;
    private final Collection<Class<? extends Agent>> classes;
    private final String goodName;

    /**
     * By population. Can be set to print AVERAGE, rather than TOTAL.
     */
    public GoodTrackerDataPrinter(String goodName, boolean printAvg) {
        this.goodName = goodName;
        this.avg = printAvg;
        this.classes = null;
    }

    /**
     * By classes. Can be set to print AVERAGE, rather than TOTAL.
     */
    public GoodTrackerDataPrinter(String goodName,
            Collection<Class<? extends Agent>> classes, boolean printAvg) {
        this.goodName = goodName;
        this.avg = printAvg;
        this.classes = classes;
    }

    /**
     * By class
     */
    public GoodTrackerDataPrinter(String goodName,
            Class<? extends Agent> agentClass, boolean printAvg) {
        this.goodName = goodName;
        Collection<Class<? extends Agent>> classes = new ArrayList<Class<? extends Agent>>();
        classes.add(agentClass);
        this.avg = printAvg;
        this.classes = classes;
    }

    /**
     * Calls this.getData(economy)
     */
    public String getCleanupData(DataOutputEconomy economy) {
        return this.getData(economy);
    }

    /**
     * Returns data based on constructor parameters. If the Agent does not 
     * have the Good, simply returns NaN.
     */
    public String getData(DataOutputEconomy economy) {
        try {
            int count = 0;
            double amount = 0.0;
            if (this.classes != null) // by class(es)
            {
                for (Agent a : economy.getAgents()) {
                    if (!this.classes.contains(a.getClass()))
                        continue;
                    count++;
                    amount += a.getGood(this.goodName).getQuantity();
                }
            } else // by population
            {
                for (Agent a : economy.getAgents()) {
                    count++;
                    amount += a.getGood(this.goodName).getQuantity();
                }
            }
            if (this.avg)
                return (amount / (double) count) + "";
            return amount + "";
        } catch (Exception ex) {
            return Double.NaN + "";
        }
    }

    /**
     * Title based on constructor parameters
     */
    public String getTitle(DataOutputEconomy economy) {
        String title;
        if (this.avg)
            title = "Avg ";
        else
            title = "Total ";
        if (this.classes != null) {
            if (this.classes.size() == 1) {
                title += this.goodName
                        + " of "
                        + Strings.getShortNames((Class<?>) this.classes
                                .toArray()[0]) + "s";
            } else {
                title += this.goodName + " of classes: ";
                for (Class<? extends Agent> cls : this.classes) {
                    title += Strings.getShortNames(cls) + "; ";
                }
            }
        } else {
            title += "quantity of " + this.goodName;
        }
        return title;
    }

    /**
     * True.
     */
    public boolean isCSVCompatible() {
        return true;
    }

    /**
     * True.
     */
    public boolean isMultiRunCompatible() {
        return true;
    }

    /**
     * Returns a Collection of GoodTrackers. A GoodTracker is included for total
     * and average quantities for each Good for each class, as well as for the
     * total quantity of each Good in the simulation.
     */
    public static Collection<DataPrinter> getDataPrintersFor(
            DataOutputEconomy economy) {
        Set<Class<? extends Agent>> seenAlready = new HashSet<Class<? extends Agent>>();
        Set<String> goods = new HashSet<String>();
        ArrayList<DataPrinter> list = new ArrayList<DataPrinter>();

        for (Agent agent : economy.getAgents()) {
            if (seenAlready.contains(agent.getClass()))
                continue;
            seenAlready.add(agent.getClass());
            for (Good good : agent.getGoods()) {
                goods.add(good.getName());
                list.add(new GoodTrackerDataPrinter(good.getName(), agent
                        .getClass(), true));
                list.add(new GoodTrackerDataPrinter(good.getName(), agent
                        .getClass(), false));
            }
        }
        for (String goodName : goods) {
            list.add(new GoodTrackerDataPrinter(goodName, false));
        }

        return list;
    }

}
