/**
 * 
 */
package eos.models;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

import eos.framework.Agent;
import eos.framework.Economy;
import eos.framework.Good;
import eos.utilities.Errors;

/**
 * A model class for the baseline
 * 
 * @author Michael Adelson
 *
 */
public class BaselineModel {

    /**
     * This Good implementation defines a fixed set of Goods for use in
     * a BaseLine Economy.
     * 
     * @author Michael Adelson
     */
    public static class BaselineGood extends NameAndQuantityGood {

        /**
         * Allowed Goods
         */
        public static String Food = "Food", Labor = "Labor", Money = "Money";

        /**
         * Register allowed Goods with superclass.
         */
        static {
            NameAndQuantityGood.registerName(BaselineGood.class, Food);
            NameAndQuantityGood.registerName(BaselineGood.class, Labor);
            NameAndQuantityGood.registerName(BaselineGood.class, Money);
        }
        
        /**
         * Construct a BaseLineGood with name.
         */
        public BaselineGood(String name) {
            super(name);
        }
    }
    
    /**
     * Abstract class defining which goods are used by Agents in the baseline
     * economy.
     * 
     * @author Mike Adelson
     */
    public static abstract class BaselineAgent extends Agent {

        // This Agent's goods
        protected BaselineGood food, labor, money;

        /**
         * Construct a BaseLineAgent
         */
        public BaselineAgent(Economy economy, Map<String, Good> goods) {
            super(economy, goods);

            if (goods.values().size() != 3)
                Errors.err("BaseLineAgent requires exactly 3 goods");
            this.food = BaselineGood.class.cast(goods
                    .get(BaselineGood.Food));
            this.labor = BaselineGood.class.cast(goods
                    .get(BaselineGood.Labor));
            this.money = BaselineGood.class.cast(goods
                    .get(BaselineGood.Money));
        }

        /**
         * Returns a List of food, labor, money.
         */
        public Collection<Good> getGoods() {
            Collection<Good> goods = new ArrayList<Good>();
            goods.add(this.food);
            goods.add(this.labor);
            goods.add(this.money);

            return goods;
        }

        /**
         * Returns food, labor, or money, or throws an exception.
         */
        public Good getGood(String goodName) {
            if (goodName.equals(BaselineGood.Money))
                return this.money;
            if (goodName.equals(BaselineGood.Food))
                return this.food;
            if (goodName.equals(BaselineGood.Labor))
                return this.labor;
            throw new RuntimeException("Agent does not have Good " + goodName);
        }

        /**
         * Returns the Agent's quantity of money.
         */
        public double getNetWorth() {
            return this.money.getQuantity();
        }

    }

    /**
     * Abstract firm class for baseline
     * 
     * @author Michael Adelson
     *
     */
    public static abstract class BaselineFirm extends BaselineAgent {

        /**
         * Create a new Firm in economy.
         */
        public BaselineFirm(Economy economy, Map<String, Good> goods) {
            super(economy, goods);
        }

        /**
         * Firms are a more specific type of agent in that they can convert labor
         * into other goods and lose any unused labor at the end of each timestep.
         */
        public final void act() {
            produce();
            discardLabor();
            checkIfBankrupt();
            bid();
        }

        /**
         * Check if firm is bankrupt; the firm should die if it is in fact bankrupt
         */
        abstract protected void checkIfBankrupt();

        /**
         * bid on labor and sell goods
         */
        abstract protected void bid();

        /**
         * Set the amount of labor that this firm has to 0
         */
        private void discardLabor() {
            labor.withdraw(labor.getQuantity() + 1.0);
        }

        /**
         * Convert labor into another good
         */
        abstract protected void produce();

        /**
         * Returns a double value representing this Firm's net worth in terms of
         * money on hand
         */
        public double getNetWorth() {
            return money.getQuantity();
        }

    }
    
    /**
     * Abstract farm class for the Baseline
     * 
     * @author Michael Adelson
     *
     */
    public static abstract class BaselineFarm extends BaselineFirm {

        /**
         * variables used in the production function
         */
        protected final double A = 5; //must be > 0
        protected final double B = 2; //must be > 0
        
        /**
         * Create a new Firm in economy.
         */
        public BaselineFarm(Economy economy, Map<String, Good> goods) {     
            super(economy, goods);
            if (A <= 0) {
                throw new RuntimeException("A in the production is function must be > 0. Right now it is " + A + ".");
            }
            if (B <= 1) {
                throw new RuntimeException("B in the production is function must be > 1. Right now it is " + B + ".");
            }
          
        }
        
        /**
         * bid on labor and sell goods
         */
        abstract protected void bid();
        
        /**
         * if the farm has no money, no food, and no labor then it has nothing of value
         * and should die()
         */
        protected void checkIfBankrupt() {      
            if (money.getQuantity() <= 0.0 && food.getQuantity() <= 0.0 && labor.getQuantity() <= 0.0) {
                die();
            }       
        }
        
        /**
         * Convert labor into Food
         */
        protected final void produce() {
            food.deposit(convertToFood(labor.withdraw(labor.getQuantity())));
        }
        
        /**
         * Returns the amount of food that the given amount of labor can
         * produce according to the production function:
         * Food = A * labor ^ (1 / B)
         */
        protected final double convertToFood(double laborQuantity) {
            //return A * Math.pow(laborQuantity, 1 / B);
            return Math.min(2.5 * laborQuantity, 60);
            //return 20 * laborQuantity / (10 + 0.2 * laborQuantity);
        }
        
    }

    /**
     * Abstract laborer class for baseline
     * 
     * @author Michael Adelson
     *
     */
    public static abstract class BaselineLaborer extends BaselineAgent{
        
        /**
         * The amount of food that the agent eats on every timestep
         */
        protected static final double EAT_AMOUNT = 1.0;
        
        /**
         * The amount of labor that the agent gets on every timestep
         */
        protected static final double LABOR_ALLOWANCE = 1.0;
        
        /**
         * Create a new Laborer in economy.
         */
        public BaselineLaborer(Economy economy, Map<String, Good> goods) {
            super(economy, goods);       
        }
        
        /**
         * Laborers are more specific agents in that they must eat on every timestep
         * and they update their labor an every timestep to a certain number of units
         */
        public final void act() {
            
            /**
             * The agent must be alive in order to act
             */
            if (!isAlive()) {
                throw new RuntimeException("This agent is dead and therefore should not act().");
            }
            
            /**
             *  If the agent can't eat, then it shouldn't be able to do anything else
             */
            if (!eat()) {
                return;
            }
            
            updateLabor();
            
            bid();
            
        }
        
        /**
         * Should be extended by a subclass; this is where the laborer sells his labor to firms on
         * the labor market and sells/buys food on the food market
         */
        abstract protected void bid();
        
        /**
         * sets the amount of labor that this laborer has to LABOR_ALLOWANCE
         */
        private final void updateLabor() {
            
            labor.withdraw(labor.getQuantity() + 1.0); //remove all the labor from this laborer
            labor.deposit(LABOR_ALLOWANCE);
            
        }
        
        /**
         * decrements the amount of food the agent has by EAT_AMOUNT units
         * returns if there was enough food to eat
         */
        private final boolean eat() {
            
            double amountEaten = food.withdraw(EAT_AMOUNT);
            if (amountEaten < EAT_AMOUNT) {
                die();
                return false;
            }
            return true;
            
        }
        
        /**
         * Returns a double value representing this Agent's net worth.
         */
        public double getNetWorth() {
            return money.getQuantity();
        }
        
        
    }
}
