/**
 * This class extends all of the methods in the Agent abstract class in a manner in
 * which is befitting of an agent that has labor to sell on the labor market and needs
 * to eat a certain amount of food on every timestep. The only method that needs to be
 * implemented by subclasses of this class is bid(), in which the agent sells labor and
 * sells and/or buys food
 * 
 * edit 3-23-2009 Chris Rucinski
 */

package eos.models;

import java.util.Map;
import eos.framework.*;
import eos.utilities.Errors;

/**
 * A Laborer who eats EAT_AMOUNT of food (returned by the abstract getFood()
 * method) and produces LABOR_ALLOWANCE of labor (placed in abstract getLabor())
 * each timestep.
 * 
 * @author Chris Rucinski
 * @author Michael Adelson
 */
public abstract class Laborer extends Agent {

    /**
     * The amount of food that the agent eats on every timestep
     */
    public static final double EAT_AMOUNT = 1.0;

    /**
     * The amount of labor that the agent gets on every timestep
     */
    public static final double LABOR_ALLOWANCE = 1.0;
    
    /**
     * The amount of labor sold during the last time step
     */
    private double lastLaborSold;

    /**
     * Create a new Laborer in economy.
     */
    public Laborer(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
    }

    /**
     * Calls eat(), updateLabor(), and laborerSubclassAct().
     */
    public final void act() {
        Errors.Assert(this.isAlive(), "Must be alive to act()");
        if (!this.eat()) /* If the Agent starves, it can't do anything else */
            return;
        
        this.lastLaborSold = LABOR_ALLOWANCE - this.getLabor().getQuantity();
        this.updateLabor();
        this.laborerSubclassAct();
    }
    
    /**
     * Implemented by a Subclass to perform further actions during act().
     */
    abstract protected void laborerSubclassAct();

    /**
     * sets the amount of labor that this laborer has to LABOR_ALLOWANCE
     */
    protected void updateLabor() {
        // remove all Labor first.
        this.getLabor().withdraw(this.getLabor().getQuantity() + 1.0); 
        this.getLabor().deposit(LABOR_ALLOWANCE);
    }

    /**
     * If the Laborer cannot eat EAT_AMOUNT, die and return false. Otherwise,
     * return true.
     */
    protected boolean eat() {
        if (this.getFood().withdraw(EAT_AMOUNT) < EAT_AMOUNT) {
            this.die();
            return false;
        }
        return true;
    }
    
    /**
     * @return the amount of labor sold during the last time step.
     */
    protected final double getLastLaborSold() {
        return this.lastLaborSold;
    }

    /**
     * Should be implemented to return a Good representing the Laborer's Food.
     */
    abstract protected Good getFood();

    /**
     * Should be implemented to return a Good representing the Laborer's labor.
     */
    abstract protected Good getLabor();
}
