/**
 * 
 */
package eos.models;

import java.util.*;

import eos.utilities.Errors;
import eos.utilities.Strings;

/**
 * A Continuous Good with a Name and Quantity. Implementations should statically
 * register legal names for Goods in a static { } block using the method
 * registerName().
 * 
 * NameAndQuantityGood makes it easy for EOS model creators to define a Good
 * class for all NameAndQuantityGoods they intend to use. It further allows for
 * dynamic addition of new allowed Goods as the simulation progresses if the
 * programmer so desires. Generally though, this class makes it easy to get good
 * error checking with minimal coding. See BaseLine2Good for an example of a
 * class that extends NameAndQuantityGood.
 * 
 * @author Michael Adelson
 */
public class NameAndQuantityGood extends ContinuousGood {

    /**
     * A static Map of extending classes to registered names. This is used to
     * check if a given name has been registered for use by the Good. For
     * example, if a BaseLineGood, which registers Food, Money, and Labor were
     * to be instantiated with "Chocolate", NameAndQuantityGood would throw an
     * exception.
     */
    private static final Map<Class<? extends NameAndQuantityGood>, Collection<String>> registeredNames;

    /**
     * Initialize registeredNames.
     */
    static {
        registeredNames = new HashMap<Class<? extends NameAndQuantityGood>, Collection<String>>();
    }

    /**
     * The name of this Good.
     */
    protected final String name;

    /**
     * Construct a new NameAndQuantityGood with name. Throws an exception if
     * name is not registered for the extending class. Checks are performed in a
     * case-insensitive fashion, but the name will be stored as
     * Strings.capitalize(name).
     */
    public NameAndQuantityGood(String name) {
        super();

        this.name = Strings.capitalize(name);

        // get the registered names for this class of Good.
        Collection<String> names = registeredNames.get(this.getClass());

        if (names == null)
            Errors.err("Class " + this.getClass().getName()
                    + " did not register itself with NameAndQuantityGood");
        if (!names.contains(this.name))
            Errors.err("Class " + this.getClass().getName()
                    + " did not register name " + name
                    + " with NameAndQuantityGood");
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.framework.Good#getName()
     */
    public String getName() {
        return this.name;
    }

    /**
     * Registers a capitalized version of name for the registering class. This
     * method should be called within a static { } block in an extending class.
     **/
    public static void registerName(
            Class<? extends NameAndQuantityGood> registeringClass, String name) {
        if (!registeredNames.containsKey(registeringClass))
            registeredNames.put(registeringClass, new ArrayList<String>(1));

        String capName = Strings.capitalize(name);
        registeredNames.get(registeringClass).add(capName);
    }

}
