/**
 * 
 */
package eos.models;

import java.util.*;
import eos.framework.*;
import eos.models.SingleOwnerFirms.*;
import eos.utilities.Errors;
import eos.utilities.Simulations;

/**
 * Subclass of Firm which implements the SingleOwnerFirm interface. Act()
 * resolves transfers and shutdowns, then calls ownedFirmSubclassAct()
 * 
 * @author Michael Adelson
 */
public abstract class OwnedFirm extends Firm implements SingleOwnerFirm {

    // the Firm's owner
    private SingleOwnerFirmOwner owner = null;

    // is the Firm to be shutDown?
    private boolean isShuttingDown = false;

    // is the Firm activated?
    protected boolean isActivated;

    // outstanding transfers
    protected final HashMap<String, Double> transfers;

    /**
     * A basic static factory method for OwnedFirms.
     * 
     * @param founder
     *            the Firm's first owner (non-null)
     * @param firmClass
     *            the class of the Firm
     * @return a new, non-activated Firm with 0 of each of the founder's Goods.
     */
    public static <T extends OwnedFirm> T createFor(Economy economy,
            SingleOwnerFirmOwner founder, Class<T> firmClass) {
        Errors.Assert(founder != null, "Founder must be non null");

        // create Goods -- do not include any Firms
        Map<String, Good> goods = new HashMap<String, Good>();
        for (Good good : founder.asAgent().getGoods()) {
            if (good instanceof SingleOwnerFirm)
                continue;
            Good firmGood = Simulations.createGood(good.getClass(), good
                    .getName());
            firmGood.withdraw(firmGood.getQuantity()); // set quantity to 0
            goods.put(firmGood.getName(), firmGood);
        }

        // instantiate the firm
        T firm = Simulations.create(firmClass, economy, goods);

        // deactivate
        firm.isActivated = false;

        // set owner
        firm.transferOwnership(founder);

        return firm;
    }

    /**
     * Construct an OwnedFirm. This constructor should ONLY be called by the
     * simulation class. Owners wishing to create new Firms should call that
     * Firm's static factory method if available, or OwnedFirm.createFor().
     * Firms built with this constructor begin as activated. Factory methods for
     * specific Firms which use this constructor should be sure to set
     * firm.isActivated to false before returning the new Firm.
     **/
    public OwnedFirm(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
        this.transfers = new HashMap<String, Double>();
        this.isActivated = true; // starts as activated.
    }

    /**
     * Resolve transfers and shutDowns. Calls the abstract method
     * ownedFirmSubclassAct();
     */
    protected final void firmSubclassAct() {
        Errors.Assert(this.isActivated, "Firm must be activated to act()");
        this.resolveTransfers();
        if (this.isShuttingDown) {
            this.resolveShutDown();
            return;
        }
        this.ownedFirmSubclassAct();
    }

    /**
     * Must be implemented by subclasses.
     */
    abstract protected void ownedFirmSubclassAct();

    /**
     * Fulfill all transfers to the extent possible, removing completed
     * transfers.
     */
    protected void resolveTransfers() {
        if (this.transfers.isEmpty())
            return;

        // create a seperate list to iterate over to avoid
        // ConcurrentModificationExceptions
        List<String> goodNames = new ArrayList<String>(this.transfers.size());
        goodNames.addAll(this.transfers.keySet());
        
        for (String goodName : goodNames) {
            double amountOwed = this.transfers.get(goodName);
            double amountTransferred = this.transferToOwner(goodName,
                    amountOwed);
            if (amountTransferred < amountOwed) // reduce the amount owed
                this.transfers.put(goodName, amountOwed - amountTransferred);
            else
                // remove the transfer
                this.transfers.remove(goodName);
        }
    }

    /**
     * Shut down this Firm, transferring all Goods possible to Owner. Can be
     * overriden. Assumes that labor has been discarded.
     */
    protected void resolveShutDown() {
        for (Good good : this.getGoods()) {
            if (good.getQuantity() > 0.0)
                this.transferToOwner(good.getName(), good.getQuantity());
        }

        this.die();
    }

    /**
     * Calls isBankrupt() (not implemented by OwnedFirm) to determine if the
     * Firm is Bankrupt, and calls shutDown() if isBankrupt() returns true.
     */
    protected void checkIfBankrupt() {
        if (this.isBankrupt())
            this.shutDown();
    }

    // SingleOwnerFirm methods

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#asAgent()
     */
    public Agent asAgent() {
        return this;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#getOutstandingTransfer
     * (java.lang.String)
     */
    public double getOutstandingTransfer(String goodName) {
        Double amount = this.transfers.get(goodName);
        if (amount != null)
            return amount;
        return 0.0;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#getOwner()
     */
    public SingleOwnerFirmOwner getOwner() {
        return this.owner;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#getQuantity(java.lang
     * .String)
     */
    public double getQuantity(String goodName) {
        return SingleOwnerFirms.getQuantity(this);
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#liquidate()
     */
    public void shutDown() {
        this.isShuttingDown = true;
    }

    /**
     * Activates the Firm. This implementation is designed such that only Firms
     * created with that Firm's static factory method (or OwnedFirm.createFor())
     * need to be activated. This means that only Firms with an owner that are
     * not initially placed in the economy require activation. If a Firm acts()
     * without having activated, an error is thrown. If this method is called
     * and the firm is already activated, it simply returns true. This
     * implementation calls simply transfers all costs as Goods to the Firm if
     * possible, if the Firm lacks those Goods the cost is simply withdrawn from
     * the owner and discarded. One simple way for subclasses to change this
     * behavior is to override activate() with a method that calls
     * super.activate(), and then does more with the received Goods if
     * super.activate() returns true.
     */
    public boolean activate() {
        Errors.Assert(this.owner != null, "Must have an owner to activate");
        if (this.isActivated)
            return true;

        Map<String, Double> costs = this.getRemainingActivationCosts();
        boolean hasSufficientGoods = true;
        for (String goodName : costs.keySet())
            // if the owner has less of goodName than the cost...
            if (this.owner.asAgent().getGood(goodName).getQuantity() < costs
                    .get(goodName)) {
                hasSufficientGoods = false;
                break;
            }

        if (!hasSufficientGoods) // can't activate
            return false;

        // Pay all costs
        for (String goodName : costs.keySet())
            this.owner.asAgent().getGood(goodName)
                    .withdraw(costs.get(goodName));

        // transfer to Firm if possible
        for (Good good : this.getGoods()) { // for each of this Firm's goods
            if (costs.containsKey(good.getName())) // if it is a cost
                good.deposit(costs.get(good.getName())); // deposit the cost
        }

        // activated!
        this.isActivated = true;
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#requestTransfer(java.
     * lang.String, double)
     */
    public void requestTransfer(String goodName, double amount) {
        Errors.Assert(amount >= 0.0, "Amount must be >= 0.0");
        Double oldAmount = this.transfers.get(goodName);
        if (oldAmount != null)
            this.transfers.put(goodName, amount + oldAmount);
        else
            this.transfers.put(goodName, amount);
    }

    /**
     * A simple implementation: Makes all outstanding transfers possible, then
     * transfers ownership of the firm. This does not guarantee that outstanding
     * transfers are fulfilled before ownership is transferred. However, it does
     * guarantee that there are no outstanding transfers once the Firm changes
     * owners (they have all been fulfilled or relinquished). If the Owner is
     * null, newOwner is simply given ownership of the Firm. Warns if the new
     * and old owners are the same.
     */
    public void transferOwnership(SingleOwnerFirmOwner newOwner) {
        Errors
                .warnIf(this.owner == newOwner,
                        "New and old owners are the same");

        if (this.owner != null) {
            // try to resolve outstanding transfers
            for (Good good : this.getGoods()) {
                double amountOwed = this.getOutstandingTransfer(good.getName());
                if (amountOwed > 0.0) { // if a transfer is outstanding
                    // try to resolve
                    this.transferToOwner(good.getName(), amountOwed);
                    // relinquish the transfer if necessary
                    this.relinquishTransfer(good.getName());
                }
            }

            // give up ownership
            this.owner.relinquishOwnership(this);
        }

        // transfer ownership of the Firm
        newOwner.assumeOwnership(this);
        this.owner = newOwner;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#transferToFirm(java.lang
     * .String, double)
     */
    public double transferToFirm(String goodName, double amount) {
        return SingleOwnerFirms.transferToFirm(this, goodName, amount);
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#transferToOwner(java.
     * lang.String, double)
     */
    public double transferToOwner(String goodName, double amount) {
        return SingleOwnerFirms.transferToOwner(this, goodName, amount);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * eos.models.SingleOwnerFirms.SingleOwnerFirm#relinquishTransfer(java.lang
     * .String)
     */
    public void relinquishTransfer(String goodName) {
        this.transfers.remove(goodName);
    }

    /**
     * Throws an Exception.
     */
    public void deposit(double amount) {
        SingleOwnerFirms.deposit(this, amount);
    }

    /**
     * The name of this Firm is it's shortened class name.
     */
    public String getName() {
        return SingleOwnerFirms.getName(this);
    }

    /**
     * Returns 1.
     */
    public double getQuantity() {
        return SingleOwnerFirms.getQuantity(this);
    }

    /**
     * True;
     */
    public boolean isDiscrete() {
        return SingleOwnerFirms.isDiscrete(this);
    }

    /**
     * Throws an Exception.
     */
    public double withdraw(double amount) {
        return SingleOwnerFirms.withdraw(this, amount);
    }
}