package eos.models;

import java.util.*;

import eos.framework.*;
import eos.models.SingleOwnerFirms.*;
import eos.utilities.Errors;

/**
 * Laborer that implements the SingleOwnerFirmOwner interface.
 * 
 * @author Michael Adelson
 */
public abstract class OwnerLaborer extends Laborer implements
        SingleOwnerFirmOwner {

    /**
     * This Laborer's Firms
     */
    protected ArrayList<SingleOwnerFirm> firms = null;

    /**
     * Construct an OwnerLaborer. If goods contains any SingleOwnerFirms, 
     * transfers ownership of those firms to this OwnerLaborer.
     */
    public OwnerLaborer(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
        
        for (Good good : goods.values()) {
            if (good instanceof SingleOwnerFirm)
                ((SingleOwnerFirm) good).transferOwnership(this);
        }
    }

    /**
     * Calls ownedLaborerSubclassAct()
     */
    protected final void laborerSubclassAct() {
        this.ownedLaborerSubclassAct();
    }

    /**
     * Should be implemented by subclasses to perform further actions during
     * act().
     */
    abstract protected void ownedLaborerSubclassAct();

    /**
     * The owner will never starve while his Firms could have supplied him with
     * Food. Furthermore, if the owner dies, all his firms are shut down.
     */
    protected boolean eat() {
        // if the owner has enough to eat, just eat!
        if (this.getFood().getQuantity() >= Laborer.EAT_AMOUNT)
            return super.eat();

        // else if the owner has firms, try to get food from them.
        if (this.firms != null) {
            double amount = this.getFood().withdraw(this.getFood().getQuantity());
            for (SingleOwnerFirm firm : this.firms) {
                amount += firm.transferToOwner(this.getFood().getName(),
                        2.0 * Laborer.EAT_AMOUNT);
                if (amount >= Laborer.EAT_AMOUNT) // got enough food
                    break;
            }

            // put the food in the owner's Food Good
            this.getFood().deposit(amount);

            // shut down all firms if necessary
            if (this.getFood().getQuantity() < Laborer.EAT_AMOUNT) {
                for (SingleOwnerFirm firm : this.firms)
                    firm.shutDown();
            }
        }

        // and now eat...
        return super.eat();
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirmOwner#asAgent()
     */
    public Agent asAgent() {
        return this;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirmOwner#assumeOwnership(
     * eos.models.SingleOwnerFirms.SingleOwnerFirm)
     */
    public void assumeOwnership(SingleOwnerFirm firm) {
        if (this.firms == null)
            this.firms = new ArrayList<SingleOwnerFirm>(1);
        this.firms.add(firm);
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirmOwner#getFirms()
     */
    public Collection<SingleOwnerFirm> getFirms() {
        return this.firms;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirmOwner#relinquishOwnership
     * (eos.models.SingleOwnerFirms.SingleOwnerFirm)
     */
    public void relinquishOwnership(SingleOwnerFirm firm) {
        Errors.Assert(this.firms.contains(firm), "Firm " + firm.getName()
                + " is not owned by this Agent");
        this.firms.remove(firm);
    }
}
