/**
 * 
 */
package eos.models;

import java.util.Collection;
import java.util.Map;

import eos.framework.*;
import eos.utilities.Errors;
import eos.utilities.Strings;

/**
 * Contains the interfaces for SingleOwnerFirm and SingleOwnerFirmOwner. Also
 * contains many static implementations of the interface methods to make it
 * easier for these interfaces to be rapidly implemented.
 * 
 * @author Michael Adelson
 */
public class SingleOwnerFirms {

    /*-------------------- Interfaces --------------------*/

    /**
     * An interface for a Firm owned by a single Agent. This interface extends
     * Good, allowing the Firm to be traded by a Market.
     * 
     * @author Michael Adelson
     */
    public static interface SingleOwnerFirm extends Good {

        /**
         * @return the owner of this Firm
         */
        SingleOwnerFirmOwner getOwner();

        /**
         * @return an Agent view of this Firm
         */
        Agent asAgent();

        /**
         * @return this Firm's quantity of goodName.
         */
        double getQuantity(String goodName);

        /**
         * Transfer amount of Good goodName to Owner if possible. Return the
         * amount transferred. Unlike requestTransfer(), a call to this method
         * which fails to return the entire amount will not make up for it
         * later.
         */
        double transferToOwner(String goodName, double amount);

        /**
         * Request that amount of Good goodName be transferred to Owner. This
         * transfer should happen as soon as possible, but not in the current
         * time step. This prevents the problem of Owners disrupting their
         * Firm's activities by withdrawing funds before the Firm acts. This
         * also avoids the problem of the Firm acting before it's owner. In this
         * case, the Firm might spend all funds, effectively denying the owner a
         * chance to withdraw them.
         */
        void requestTransfer(String goodName, double amount);

        /**
         * Relinquish any outstanding transfers for goodName. After a call to
         * this method, getOutstandingTransfer(goodName) should return 0.
         */
        void relinquishTransfer(String goodName);

        /**
         * @return the amount of Good goodName which has been requested but not
         *         yet transferred to the Owner.
         **/
        double getOutstandingTransfer(String goodName);

        /**
         * Transfer the ownership of this Firm to newOwner.
         */
        void transferOwnership(SingleOwnerFirmOwner newOwner);

        /**
         * Is the Firm bankrupt?
         */
        boolean isBankrupt();

        /**
         * Transfer amount of goodName from the Owner to the Firm if possible.
         * Return the amount transferred. This method can also be used to
         * relinquish requested transfers, but relinquishTransfer is preferable.
         */
        double transferToFirm(String goodName, double amount);

        /**
         * Bankrupt this Firm, transferring all Goods to Owner if possible. The
         * liquidation might not actually happen until the next time step.
         */
        void shutDown();

        /**
         * Activate this Firm. In implemenations where Agents can start new
         * Firms, this method must be called in order to get the Firm going.
         * This is where the Owner must pay any costs associated with starting
         * the Firm. These costs are returned by getRemainingActivationCosts().
         * Returns true if the Firm successfully activates, false otherwise. The
         * simplest implementation is to have the false case not withdraw any
         * Goods from the Owner. However, if activation should require multiple
         * days worth of labor, for example, it may be preferable to have each
         * activation of this method reduce the costs returned by
         * getRemainingActivationCosts().
         */
        boolean activate();

        /**
         * Get the remaining activation costs for this Firm.
         * 
         * @return a Mapping of Good names to double costs. The Map should be
         *         empty if the Firm is activated.
         */
        Map<String, Double> getRemainingActivationCosts();
    }

    /**
     * An interface for an Owner of a SingleOwnerFirm. Can own multiple Firms.
     * 
     * @author Michael Adelson
     */
    public static interface SingleOwnerFirmOwner {

        /**
         * @return an Agent view of this Owner.
         */
        Agent asAgent();

        /**
         * @return a Collection view of this Owner's Firms.
         */
        Collection<SingleOwnerFirm> getFirms();

        /**
         * Assume ownership of Firm. This should be called by
         * SingleOwnerFirm.transferOwnership().
         */
        void assumeOwnership(SingleOwnerFirm firm);

        /**
         * Relinquish ownership of Firm. This should be called by
         * SingleOwnerFirm.transferOwnership().
         */
        void relinquishOwnership(SingleOwnerFirm firm);
    }

    /*-------------------- Static Methods --------------------*/

    /*
     * These method provide basic implementations of many of the like-named
     * interface methods. They may be called by implementing classes that do not
     * wish to write their own implementations.
     */

    /*
     * (non-Javadoc)
     * 
     * @see eos.models.SingleOwnerFirms.SingleOwnerFirm#getQuantity(java.lang
     * .String)
     */
    public static double getQuantity(SingleOwnerFirm self, String goodName) {
        return self.asAgent().getGood(goodName).getQuantity();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * eos.models.SingleOwnerFirms.SingleOwnerFirm#transferToFirm(java.lang.
     * String, double)
     */
    public static double transferToFirm(SingleOwnerFirm self, String goodName,
            double amount) {
        Errors.Assert(amount >= 0.0, "Amount must be >= 0");
        Errors.Assert(self.getOwner() != null, "Owner must not be null");

        double actualAmount = self.getOwner().asAgent().getGood(goodName)
                .withdraw(amount);
        self.asAgent().getGood(goodName).deposit(actualAmount);

        return actualAmount;
    }

    /**
     * If this firm has an owner, transfer up to amount of goodName to owner.
     * 
     * @return the amount actually transferred.
     */
    public static double transferToOwner(SingleOwnerFirm self, String goodName,
            double amount) {
        Errors.Assert(amount >= 0.0, "Amount must be >= 0.0");
        if (self.getOwner() == null) // no owner case.
            return 0.0;
        double actualAmount = self.asAgent().getGood(goodName).withdraw(amount);
        self.getOwner().asAgent().getGood(goodName).deposit(actualAmount);

        return actualAmount;
    }

    /**
     * Throws an Exception.
     */
    public static void deposit(SingleOwnerFirm self, double amount) {
        throw new RuntimeException("This method should not be used for "
                + self.getClass());
    }

    /**
     * The name of this Firm is it's shortened class name + the Firms Agent ID.
     * This gives every Firm a unique, meaningful name.
     */
    public static String getName(SingleOwnerFirm self) {
        return Strings.getShortNames(self.getClass()) + self.asAgent().getID();
    }

    /**
     * Returns 1.
     */
    public static double getQuantity(SingleOwnerFirm self) {
        return 1;
    }

    /**
     * True;
     */
    public static boolean isDiscrete(SingleOwnerFirm self) {
        return true;
    }

    /**
     * Throws an Exception.
     */
    public static double withdraw(SingleOwnerFirm self, double amount) {
        throw new RuntimeException("This method should not be used for "
                + self.getClass());
    }

    /*-------------------- Constructor --------------------*/

    /**
     * Non-instantiable.
     */
    private SingleOwnerFirms() {
    }
}
