/**
 * 
 */
package eos.utilities;

import eos.agents.SimpleLaborer;
import eos.economies.BL321Economy;
import eos.framework.*;
import eos.markets.CallAuctionMarket;
import eos.models.BaselineModel;

import java.lang.reflect.Constructor;
import java.util.*;

/**
 * Utilities for creating simulation classes.
 * 
 * @author Michael Adelson
 */
public class Simulations {

    // Statics

    /**
     * Instantiates an Object, trying all constructors possible.
     * 
     * @param cls
     *            the class of object to create.
     * @param pArgs
     *            the paramaters to pass to the constructor. If null/empty this
     *            method uses cls.newInstance().
     */
    public static <T> T create(Class<T> cls, Object... pArgs) {
        if (pArgs == null || pArgs.length == 0) {
            try {
                return cls.newInstance();
            } catch (Exception ex) {
                Errors.err("Could not instantiate " + cls.getName() + ": "
                        + ex.getMessage());
                return null;
            }
        } else {
            Constructor<?>[] ctors = cls.getDeclaredConstructors();
            if (ctors.length == 0) {
                ctors = cls.getConstructors();
                Errors.Assert(ctors.length != 0,
                        "Could not find constructors for class: "
                                + cls.getName());
            }

            Exception ex = null;
            for (Constructor<?> ctor : ctors) {
                try {
                    return cls.cast(ctor.newInstance(pArgs));
                } catch (Exception ex0) {
                    ex = ex0;
                }
            }
            Errors.err(ex, "Could not instantiate " + cls.getName()
                    + " with parameters " + Arrays.toString(pArgs));
            return null;
        }
    }

    /**
     * @param marketClass
     *            the Market's class
     * @param currencyName
     *            the Market's currency
     * @param productName
     *            the Market's product
     * @param economy
     *            the enclosing Economy
     * @return the instantiated Market
     */
    public static Market createMarket(Class<? extends Market> marketClass,
            String currencyName, String productName, Economy economy) {
        return create(marketClass, economy, currencyName, productName);
    }

    /**
     * 
     * @param cls
     *            the class of the Agent
     * @param goodNamesAndClasses
     *            a map of good names to the classes used to instantiate them
     * @param startingQuantities
     *            the starting quantities of the Agent's Goods
     * @param economy
     *            the enclosing Economy
     * @param numAgents
     *            the number of Agents to create
     * @return an array (length = numAgents) of instantiated Agents
     */
    public static <A extends Agent> List<A> createAgents(Class<A> cls,
            LinkedHashMap<String, Class<? extends Good>> goodNamesAndClasses,
            double[] startingQuantities, Economy economy, int numAgents) {
        ArrayList<A> agents = new ArrayList<A>(numAgents);
        for (int i = 0; i < numAgents; i++)
            agents.add(createAgent(cls, goodNamesAndClasses,
                    startingQuantities, economy));
        return agents;
    }

    /**
     * 
     * @param cls
     *            the class of the Agent
     * @param goodNamesAndClasses
     *            a map of good names to the classes used to instantiate them
     * @param startingQuantities
     *            the starting quantities of the Agent's Goods
     * @param economy
     *            the enclosing Economy
     * @return the instantiated Agent
     */
    public static <A extends Agent, G extends Good> A createAgent(Class<A> cls,
            LinkedHashMap<String, Class<? extends Good>> goodNamesAndClasses,
            double[] startingQuantities, Economy economy) {

        Errors.Assert(startingQuantities.length == goodNamesAndClasses.size());

        Map<String, Good> map = new HashMap<String, Good>();
        int index = 0;
        for (Map.Entry<String, Class<? extends Good>> e : goodNamesAndClasses
                .entrySet()) {
            Good good = createGood(e.getValue(), e.getKey());
            good.deposit(startingQuantities[index++]);
            map.put(e.getKey(), good);
        }

        return create(cls, economy, map);
    }

    /**
     * Instantiate a Good object with these parameters
     * 
     * @param goodClass
     *            the class of Good to be instantiated
     * @param name
     *            the name of the Good to be instantiated
     */
    public static <G extends Good> G createGood(Class<G> goodClass, String name) {
        return create(goodClass, name);
    }

    /**
     * Returns a Map that can be used as an argument to createAgent(s)().
     * 
     * @param goodClass
     *            the Good class used to instantiate the Goods
     * @param names
     *            the names of the Goods
     * @return a new LinkedHashMap of names to classes.
     */
    public static NameClsMap getMap(Class<? extends Good> goodClass,
            String... names) {
        NameClsMap map = new NameClsMap();
        for (String name : names)
            map.put(name, goodClass);

        return map;
    }

    /**
     * Returns a Map that can be used as an argument to createAgent(s)().
     * 
     * @param namesAndClasses
     *            a list of all Good names and Classes used. Matches the first
     *            name to the first class, the second name to the second class,
     *            etc. If only 1 class appears, maps all names to that class.
     *            Otherwise, if the number of names != number of classes, throws
     *            an exception.
     * @return A new LinkedHashMap of names to classes
     */
    @SuppressWarnings("unchecked")
    // unchecked cast to Class<? extends Good>
    public static NameClsMap getMap(Object... namesAndClasses) {
        NameClsMap map = new NameClsMap();

        // return an empty map
        if (namesAndClasses.length == 0)
            return map;

        Queue<String> names = new LinkedList<String>();
        Queue<Class<? extends Good>> classes = new LinkedList<Class<? extends Good>>();

        // extract names and classes in the correct order
        for (Object obj : namesAndClasses) {
            if (obj instanceof String)
                names.add((String) obj);
            else
                try {
                    classes.add((Class<? extends Good>) obj);
                } catch (Exception ex) {
                    Errors
                            .err(ex,
                                    "Arguments must either be Strings or a Class<? extends Good>'s");
                }
        }

        // check for all to one case
        if (classes.size() == 1 && names.size() > 0) {
            for (String name : names)
                map.put(name, classes.peek());
            return map;
        }

        // check that classes and names are the same size
        Errors
                .Assert(
                        classes.size() == names.size(),
                        "Number of classes and names must be equal, or there must be n names and 1 class");

        // create the Map
        for (String name : names)
            map.put(name, classes.remove());

        return map;
    }

    /**
     * Returns a Map that can be used as an argument to createAgent(s)(). Note
     * that this does not record the quantities of the Goods, only their names
     * and classes.
     * 
     * @param goods
     *            a collection of Goods
     * @return A new LinkedHashMap of names to classes
     */
    public static NameClsMap getMap(Collection<Good> goods) {
        NameClsMap map = new NameClsMap();

        for (Good good : goods)
            map.put(good.getName(), good.getClass());

        return map;
    }

    /*---------------------- CLASSES -----------------------------------------*/

    /**
     * A Map of Good names to Classes generated by the getMap() functions and
     * used by createAgent(s)(). This is a LinkedHashMap<String, Class<? extends
     * Good>, and thus saves lots of typing. The linked map is used for
     * guaranteed iteration ordering, which is necessary for it to be used
     * parallel to a quantity array.
     * 
     * @author Michael Adelson
     */
    public static class NameClsMap extends
            LinkedHashMap<String, Class<? extends Good>> {
        /**
         * Default
         */
        private static final long serialVersionUID = 1L;
    }

    /*----------------- TEST CLIENT ------------------------------------------*/

    /**
     * Test client
     */
    public static void main(String[] args) {
        Economy e = new BL321Economy();

        Market market = createMarket(CallAuctionMarket.class, "Money", "Food",
                e);
        System.out.println("Created Market: " + market.getCurrency() + ", "
                + market.getProduct());

        LinkedHashMap<String, Class<? extends Good>> goods = getMap("Money",
                "Food", "Labor", BaselineModel.BaselineGood.class);
        double[] q = { 1, 2, 3 };

        List<? extends Agent> list = createAgents(SimpleLaborer.class, goods,
                q, e, 10);
        System.out.println("Created " + list.size() + " Agents");
        System.out.println("Agent 2 has "
                + list.get(2).getGood("Food").getQuantity() + " food");
        System.out.println("Agent 5 has "
                + list.get(5).getGood("Money").getQuantity() + " money");
        System.out.println("Agent 9 has "
                + list.get(9).getGood("Labor").getQuantity() + " labor");
    }
}
